from starling_sim.basemodel.agent.vehicles.service_vehicle import ServiceVehicle
from networkx.exception import NetworkXNoPath


class PublicTransportVehicle(ServiceVehicle):
    """
    This class describes a public transport vehicle that follows
    a time table often described with a GTFS.
    """

    SCHEMA = {"x-display": "hidden"}

    def __init__(self, simulation_model, agent_id, origin, seats, **kwargs):
        super().__init__(simulation_model, agent_id, origin, seats, **kwargs)

        # route id, describes the line of the current trip
        self.routeId = None

        # direction id, indicates if the trip is in the direct way (0) or not (1)
        self.directionId = None

    def loop_(self):
        # execute each trip assigned to the vehicle
        for trip_id in self.tripList:
            # set the vehicle trip id
            self.tripId = trip_id
            self.routeId, self.directionId = self.operator.get_route_and_direction_of_trip(trip_id)

            # get the planning associated to the trip
            self.planning = self.operator.trips[trip_id][1].copy()

            yield self.execute_process(self.execute_trip_())

    def execute_trip_(self):
        """
        Execute and follow the current trip.

        It is assumed that the tripId, routeId, directionId
        and planning attributes are already set.
        """

        # teleport to the trip start position
        next_trip_start = self.planning[0].position
        self.move_to_(next_trip_start, 0)

        # wait for the trip departure
        wait_time = self.planning[0].arrivalTime[self.tripId][0] - self.sim.scheduler.now()
        yield self.execute_process(self.spend_time_(wait_time))

        # follow the planning
        current_stop = self.planning[0]
        while self.planning:
            # go to next stop
            next_stop = self.planning[0]
            move_duration = next_stop.arrivalTime[self.tripId][0] - self.sim.scheduler.now()
            yield self.execute_process(
                self.public_transport_move_(self.operator, move_duration, current_stop, next_stop)
            )

            # process the next stop of the planning
            yield self.execute_process(self.process_stop_(next_stop))
            current_stop = next_stop

    # move method for public transports
    def public_transport_move_(self, operator, move_duration, from_stop, to_stop):
        """
        Move from one stop to another following the provided line shape if possible.

        If no line shape is available, execute a normal move or fly,
        depending on the operator parameters.

        :param operator:
        :param move_duration:
        :param from_stop:
        :param to_stop:
        :return:
        """

        # set the effective arrival time of the destination stop
        to_stop.effectiveArrivalTime[self.tripId] = self.sim.scheduler.now() + move_duration

        # try to use the line_shape file
        if from_stop is not None and to_stop is not None and operator.line_shapes is not None:
            # get the line shapes table
            line_shapes_table = operator.line_shapes

            # filter the table to get the right OD
            OD_shape = line_shapes_table[
                (line_shapes_table["stop_id_A"] == from_stop.id)
                & (line_shapes_table["stop_id_B"] == to_stop.id)
            ].iloc[
                1:,
            ]

            if not OD_shape.empty:
                # print(self.tripId)

                route_data = self.sim.environment.compute_shaped_route(
                    operator, OD_shape, from_stop.id, to_stop.id, move_duration
                )
                mode = self.mode
                check_dest = False

                # execute route data
                yield self.execute_process(
                    self.follow_route_data_(route_data, check_dest, mode, verb=False)
                )
                return

        # otherwise, use base methods
        if to_stop is not None:
            self.tempDestination = to_stop.position

        if self.operator.operationParameters["use_shortest_path"]:
            try:
                # compute and use shortest path
                yield self.execute_process(self.move_(duration=move_duration, verb=False))
                return
            except NetworkXNoPath:
                pass

        # compute approximate distance and cut in the environment
        yield self.execute_process(self.fly_(duration=move_duration, distance_factor=1, verb=False))

    def serve_stop(self, user_stop):
        """
        Determine if the stop should be processed by the service vehicle.

        For public transport, all users waiting at the current stop point are evaluated,
        so only those going with the right route, direction and destination stop
        should be picked up. For the dropoff, only the ones in the vehicle
        are processed.

        :param user_stop: UserStop object
        :return: boolean indicating if the stop should be processed
        """

        # get the stop request
        request = self.operator.requests[user_stop.requestId]

        # get the request trip id
        stop_trip = user_stop.trip

        # pickup stop
        if user_stop.type == user_stop.GET_REQUEST:
            # None/Nan trips are considered having good route and direction
            if isinstance(stop_trip, str):
                # check route and direction
                route_id, direction_id = self.operator.get_route_and_direction_of_trip(stop_trip)
                if not (route_id == self.routeId and direction_id == self.directionId):
                    return False

            # check that the destination stop is served
            destination_stop = self.operator.stopPoints[user_stop.twin.stopPoint]
            if destination_stop in self.planning:
                return True
            else:
                return False

        # dropoff stop
        elif user_stop.type == user_stop.PUT_REQUEST:
            return request.agent in self.occupants

    def exceeds_capacity(self, user_stop):
        # if the user expects this vehicle (ie this trip)
        # remove the stop from the stop point list and signal the failed pickup
        if isinstance(user_stop.tripId, str) and user_stop.tripId == self.tripId:
            request = self.operator.requests[user_stop.requestId]
            self.operator.stopPoints[request.dropoff.stopPoint].dropoffList.remove(request.dropoff)
            request.success = False
            request.pickupEvent_.succeed()



from starling_sim.basemodel.trace.trace import Traced
from starling_sim.basemodel.trace.events import InputEvent, WaitEvent, LeaveSimulationEvent
from starling_sim.utils.utils import SimulationError, LeavingSimulation
from starling_sim.utils.constants import DEFAULT_LEAVE, SIM_ERROR_LEAVE

import traceback
import copy
import logging


class Agent(Traced):
    """
    This class describes the basic features of a simulation agent.

    It should be extended (through several layers) to give more specific features and behaviours
    (represented by their available methods), to finally implement a concrete agent that can be
    simulated in a specific model.
    """

    SCHEMA = {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "properties": {
            "agent_id": {"type": "string", "title": "Agent ID", "description": "Unique identifier"},
            "agent_type": {
                "type": "string",
                "title": "Agent type",
                "description": "One of the model's agent types",
            },
            "mode": {
                "type": "string",
                "title": "Default network",
                "description": "Road network used by the agent",
            },
            "icon": {"type": "string", "title": "Agent icon", "description": "Display icon"},
            "duplicates": {
                "title": "Agent duplicates",
                "description": "Generate duplicates of this agent",
                "type": "integer",
                "minimum": 1,
            },
        },
        "required": ["agent_id", "agent_type", "mode", "icon"],
    }

    @classmethod
    def get_schema(cls):
        """
        Compute and return the class parameters schema.

        This is the place to add eventual schema post-processing

        :return: class schema
        """
        return cls.compute_schema()

    @classmethod
    def compute_schema(cls):
        """
        Get the json schema that specifies the class init parameters.

        The schema is generated recursively, by adding/mixing the properties
        of the current class to the schema of its parent class.

        :return: json schema of the class init parameters
        """

        # stop recursion at Agent class
        parent_class = cls.__bases__[0]
        if issubclass(parent_class, Agent):
            # start by evaluating the schema of the parent class
            schema = copy.deepcopy(parent_class.compute_schema())

            # if the current class has a specific schema, update the parent schema
            class_schema = cls.SCHEMA
            if class_schema and parent_class.SCHEMA != class_schema:
                cls.update_class_schema(schema, class_schema, cls)

        else:
            schema = cls.SCHEMA

        return schema

    def update_class_schema(base_schema, class_schema, the_class, catch_error=True):
        try:
            for keyword in class_schema.keys():
                # custom remove_props keyword, used for removing properties of parent schema
                if keyword == "remove_props":
                    for prop in class_schema["remove_props"]:
                        # remove the property from the base schema
                        del base_schema["properties"][prop]
                        # remove the property from the base schema required
                        if prop in base_schema["required"]:
                            base_schema["required"].remove(prop)

                # add the required properties to the parent schema
                elif keyword == "required":
                    for prop in class_schema["required"]:
                        if prop not in base_schema["required"]:
                            base_schema["required"].append(prop)

                # update and complete the parent schema properties
                elif keyword == "properties":
                    base_schema["properties"].update(class_schema["properties"])

                # add/replace other attributes
                else:
                    base_schema[keyword] = class_schema[keyword]

        except Exception as e:
            if catch_error:
                raise SimulationError(
                    "Schema generation of class {} failed with the following error: {}".format(
                        the_class.__name__, str(e)
                    )
                )
            else:
                raise e

    update_class_schema = staticmethod(update_class_schema)

    def __init__(self, simulation_model, agent_id, agent_type, mode, icon=None, **kwargs):
        """
        Initialize an agent with basic attributes (id, simulation model, etc).

        :param simulation_model:
        :param agent_id:
        :param agent_type: information about the agent's type
        :param mode: main mode of the agent, corresponds to a topology
        :param icon: icon used for the visualisation
        :param kwargs:
        """

        super().__init__(agent_id)

        self.sim = simulation_model
        self.type = agent_type
        self.mode = mode
        self.icon = icon

        # the agent profile contains all the additional properties of the input dict
        self.profile = kwargs

        self.name = None
        self.main_process = None
        self.current_process = None

        self.trace_event(InputEvent(self.sim.scheduler.now(), self))

    def execute_process(self, generator, store_process=False):
        """
        Create a process from the given generator and make it current process.

        The process is obtained from the given generator using the scheduler.
        It can be set as the current_process attribute in order to access it later,
        for interrupting it for instance.

        :param generator: Python generator method
        :param store_process: indicates if the process should be stored in self.current_process
        :return: Created SimPy Process object
        """

        process = self.sim.scheduler.new_process(generator)

        if store_process:
            self.current_process = process

        return process

    def spend_time_(self, duration="inf"):
        """
        Allow the simulation to flow while realising something, for the given duration.

        :param duration: wait duration, default is "inf" and leads to simulation limit
        :return: yield a timeout event with provided duration
        """

        # check that duration is integer
        if duration != "inf":
            int_duration = int(duration)

            if int_duration < 0:
                raise ValueError("An activity with negative duration was scheduled")

            if duration - int_duration != 0:
                self.log_message("Non integer time to spend : {}".format(duration), 40)
            else:
                duration = int_duration

        yield self.sim.scheduler.timeout(duration)

    def wait_(self, duration="inf", reason=None, message=""):
        """
        Wait during the given duration and trace a WaitEvent.

        :param duration: wait duration.
            Default is "inf" and leads to simulation limit.
        :param reason: reason of the wait time.
            Default is None.
        :param message: message for the wait event.
            Default is "".

        :return: yield a spend_time process with provided duration
        """

        wait_event = WaitEvent(
            time=self.sim.scheduler.now(), reason=reason, waiting_time=duration, message=message
        )

        yield self.execute_process(self.spend_time_(duration))

        self.trace_event(wait_event)

    def simpy_loop_(self):
        """
        Run the loop of the agent and catch simulation exceptions.

        This method is the main SimPy process of the agent, added to the
        SimPy environment by the DynamicInput class.
        """

        # set a default cause, to identify unmonitored leaving cases
        cause = DEFAULT_LEAVE

        # run the agent loop
        try:
            yield self.execute_process(self.loop_())

        # catch the LeavingSimulation exceptions
        except LeavingSimulation as e:
            # for simulation errors, log an specific message
            if isinstance(e, SimulationError):
                self.log_message("Encountered simulation error with message : {}".format(e), 40)
                self.log_message(traceback.print_exc(), 10)
                cause = SIM_ERROR_LEAVE

            # otherwise, log the cause description
            else:
                cause = str(e)
                cause_desc = self.sim.leaving_codes[cause]
                self.log_message("Leaving simulation: {} ({})".format(cause_desc, cause))

        # trace a LeaveSimulationEvent containing the leaving code
        event = LeaveSimulationEvent(self.sim.scheduler.now(), self, cause)
        self.trace_event(event)

        return

    def loop_(self):
        """
        This method represents the life cycle of an agent.

        The loop process will be added to the simulation environment
        after the agent creation. Agents may perform actions that have a duration.
        They must then yield other processes such as waiting or moving, and will
        resume at the end of their action.

        The loop must be extended to represent a specific behavior,
        using possibly inherited methods.
        """

        yield self.execute_process(self.spend_time_())

    def leave_simulation(self, cause):
        """
        Raise a LeavingSimulation exception to make the agent leave its loop and terminate its process.

        The cause is a leaving code that should be listed in the simulation model.
        It corresponds to a specific reason for the agent to leave the simulation.

        :param cause: leaving code
        :raises LeavingSimulation:
        """

        raise LeavingSimulation(cause)

    def simulation_error(self, description):
        """
        Raise a SimulationError exception to make the agent leave the
        simulation following a simulation error.

        This method should be called when an unwanted
        event occurs in a simulation, like saying "We shouldn't be here".

        :param description: description of the unwanted event
        :raises SimulationError:
        """

        raise SimulationError(description)

    def interrupt(self, message=""):
        """
        Interrupts the process contained in the current_process attribute.

        This process must handle the interruption exception
        """

        log_message = "Interrupted in its current process"

        if message != "":
            log_message += " with message : '{}'".format(message)

        self.log_message(log_message)

        self.sim.scheduler.interrupt(self.current_process, message=message)



"""
Starling has a precise repository structure in order to separate the
base classes of the framework, the simulation models and the simulation data.

You will find here the structure of different parts of the project.

**Data folder generation**

The :data:`~starling_sim.utils.paths.data_folder` and sub-folders are not included in the git repository,
but they can be generated using the ``-D`` (or ``--data-tree``) option of main.py

.. code-block:: bash

    python3 main.py -D

It is also generated when :ref:`importing the example scenarios<examples>`.

**********************************
Global structure of the repository
**********************************

.. code-block:: text

    starling-repo
    ├── starling_sim            # simulation modules
    |   |
    │   ├── basemodel           # base classes of the framework
    │   ├── models              # concrete models extending the basemodel
    |   ├── schemas             # json schemas
    │   └── utils               # programming utils
    |
    ├── tools                   # tools for the simulation runs
    ├── data                    # data folder
    ├── tests                   # testing
    ├── docs                    # documentation files and scripts
    ├── main.py
    ├── requirements.txt
    ├── .gitignore
    ├── CHANGELOG.md
    └── README.md

****************************
Structure of the data folder
****************************

.. code-block:: text

    data
    ├── common_inputs           # inputs shared between scenarios
    |
    ├── environment             # environment data
    │   |
    │   ├── graph_speeds        # .json files containing the graph speeds
    │   ├── gtfs_feeds          # .zip files containing the gtfs feeds
    │   └── osm_graphs          # .graphml files containing the OSM graphs
    |
    └── models                  # simulation scenarios
        |
        ├── SB_VS               # SB_VS scenarios
        |   |
        |   ├── scenario_1      # scenario_1 data
        |   |   |
        │   |   ├── inputs      # scenario_1 inputs
        │   |   └── outputs     # scenario_1 outputs
        |   |
        |   └── scenario_2      # scenario_2 data
        └── ...

The path to the data repository can be changed using the --data-folder option of main.py

.. code-block:: bash

    python3 main.py path/to/scenario/ --data-folder path/to/data_folder/

However, the structure of the data repository must remain the same. This is ensured by the functions
declared in starling_sim.utils.paths.py, that build the paths to the different folders by concatenating
the folder names in the correct order.

***************************
Structure of models package
***************************

.. code-block:: text

    starling_sim
    └── models
        ├── SB_VS             # public SB_VS model
        ├── ...               # other public models
        │
        └── MY_MODEL          # your own model
            |
            ├── model.py      # model module containing class Model(SimulationModel)
            ├── my_agent.py   # other simulation modules for your model
            └── ...

The models package is where concrete models extending the base model are placed.
It contains public models and the ones you developed.

The packages must be named with the model code, and contain a model.py module.
This module must contain a Model class extending SimulationModel of the base model.
See :ref:`create-models` for more information about the requirements for your models.
"""

import starling_sim
import os

#: path to the data folder
_DATA_FOLDER = "./data/"

#: name of the common inputs folder
COMMON_INPUTS_FOLDER = "common_inputs"

#: name of the environment folder
ENVIRONMENT_FOLDER_NAME = "environment"

#: name of the osm graphs folder
OSM_GRAPHS_FOLDER_NAME = "osm_graphs"

#: name of the graph speeds folder
GRAPH_SPEEDS_FOLDER_NAME = "graph_speeds"

#: name of the gtfs feeds folder
GTFS_FEEDS_FOLDER_NAME = "gtfs_feeds"

#: name of the schemas folder
SCHEMAS_FOLDER_NAME = "schemas"

#: name of the models folder
MODELS_FOLDER_NAME = "models"

#: name of the input folder
INPUT_FOLDER_NAME = "inputs"

#: name of the output folder
OUTPUT_FOLDER_NAME = "outputs"

#: unique parameters filename
PARAMETERS_FILENAME = "Params.json"

#: format of the model import string
_MODEL_IMPORT_FORMAT = "{starling_pkg}.models.{model_code}.model"


def data_folder():
    """
    Path to the data folder.
    """
    return _DATA_FOLDER


def common_inputs_folder():
    """
    Path to the common inputs folder
    """
    return os.path.join(data_folder(), COMMON_INPUTS_FOLDER, "")


def environment_folder():
    """
    Path to the environment folder.
    """
    return os.path.join(data_folder(), ENVIRONMENT_FOLDER_NAME, "")


def osm_graphs_folder():
    """
    Path to the osm graphs folder.
    """
    return os.path.join(environment_folder(), OSM_GRAPHS_FOLDER_NAME, "")


def graph_speeds_folder():
    """
    Path to the graph speeds folder.
    """
    return os.path.join(environment_folder(), GRAPH_SPEEDS_FOLDER_NAME, "")


def gtfs_feeds_folder():
    """
    Path to the gtfs feeds folder.
    """
    return os.path.join(environment_folder(), GTFS_FEEDS_FOLDER_NAME, "")


def models_folder():
    """
    Path to the models folder.
    """
    return os.path.join(data_folder(), MODELS_FOLDER_NAME, "")


def model_folder(model_code):
    """
    Path to the folder of the given model code.

    :param model_code: code of the model
    """
    return os.path.join(models_folder(), model_code, "")


def scenario_inputs_folder(scenario_folder):
    """
    Path to the input folder of the given scenario.

    :param scenario_folder: scenario folder path
    """
    return os.path.join(scenario_folder, INPUT_FOLDER_NAME, "")


def scenario_parameters_filepath(scenario_folder):
    """
    Path to the parameters file of the given scenario.

    :param scenario_folder: scenario folder path
    """
    return os.path.join(scenario_inputs_folder(scenario_folder), PARAMETERS_FILENAME)


def scenario_agent_input_filepath(scenario_folder, filename):
    """
    Get the path to the scenario input file (dynamic or initialisation file).

    First, look in the scenario input folder. If the file is not found,
    look in the common inputs folder. If the file is not there,
    raise a FileNotFoundError.

    :param scenario_folder: scenario folder path
    :param filename: name of the input file
    """

    # complete the file path with the input folder path
    filepath = os.path.join(scenario_inputs_folder(scenario_folder), filename)

    # if the file does not exist, look in the common inputs folder
    if not os.path.exists(filepath):
        filepath = os.path.join(common_inputs_folder(), filename)
        if not os.path.exists(filepath):
            raise FileNotFoundError(
                "Input file {} not found in scenario inputs folder "
                "or common inputs folder".format(filename)
            )

    return filepath


def scenario_outputs_folder(scenario_folder):
    """
    Path to the output folder in the given scenario folder.

    :param scenario_folder: scenario folder path
    """

    return os.path.join(scenario_folder, OUTPUT_FOLDER_NAME, "")


def starling_folder():
    """
    Path to the Starling folder.
    """
    return os.path.join(starling_sim.__path__[0], "..", "")


def schemas_folder():
    """
    Path to the schemas folder.
    """
    return os.path.join(os.path.dirname(__file__), "..", SCHEMAS_FOLDER_NAME, "")


def model_import_path(starling_pkg, model_code):
    """
    Import sequence for the given model.

    :param starling_pkg: name of the root starling package
    :param model_code: code of the model
    """
    return _MODEL_IMPORT_FORMAT.format(starling_pkg=starling_pkg, model_code=model_code)


# #: local prefix
# path_prefix = "./"
#
# #: data folder
# DATA_FOLDER = path_prefix + "data/"
#
# #: environment folder
# ENVIRONMENT_FOLDER = DATA_FOLDER + "environment/"
#
# #: default path to folder for loading and saving graphs with osmnx
# OSM_GRAPHS_FOLDER = ENVIRONMENT_FOLDER + "osm_graphs/"
#
# #: default path to folder containing graph speeds
# GRAPH_SPEEDS_FOLDER = ENVIRONMENT_FOLDER + "graph_speeds/"
#
# #: default path to folder containing gtfs feeds
# GTFS_FEEDS_FOLDER = ENVIRONMENT_FOLDER + "gtfs_feeds/"
#
# #: name of the folder containing the json schemas for the simulator
# SCHEMA_FOLDER_NAME = "schemas/"
#
# #: path to folder containing various json schemas
# SCHEMA_FOLDER = starling_sim.__path__[0] + "/../" + SCHEMA_FOLDER_NAME
#
# #: models folder
# MODELS_FOLDER = DATA_FOLDER + "models/"
#
# #: inputs folder
# INPUT_FOLDER_NAME = "inputs"
#
# #: outputs folder
# OUTPUT_FOLDER_NAME = "outputs"
#
# #: import path for non public models
# MODEL_IMPORT_PATH = "{starling_pkg}.models.{model_code}.model"

# _DATA_FOLDER = "./data/"
# _ENVIRONMENT_FOLDER_NAME = "environment"
# _OSM_GRAPHS_FOLDER_NAME = "osm_graphs"
# _GRAPH_SPEEDS_FOLDER_NAME = "graph_speeds"
# _GTFS_FEEDS_FOLDER_NAME = "gtfs_feeds"
# _SCHEMAS_FOLDER_NAME = "schemas"
# _MODELS_FOLDER_NAME = "models"
# _INPUT_FOLDER_NAME = "inputs"
# _OUTPUT_FOLDER_NAME = "outputs"
#
#
# class StarlingPath:
#
#     def __init__(self):
#
#         self._data_folder = None
#
#         self._ENVIRONMENT_FOLDER_NAME = None
#         self._OSM_GRAPHS_FOLDER_NAME = None
#         self._GRAPH_SPEEDS_FOLDER = None
#         self._GTFS_FEEDS_FOLDER = None
#         self._SCHEMA_FOLDER_NAME = None
#         self._MODELS_FOLDER_NAME = None
#         self._INPUT_FOLDER_NAME = None
#         self._OUTPUT_FOLDER_NAME = None
#
#         self._sep = None
#
#         self.set_attributes()
#
#     def set_attributes(self, data_folder=_DATA_FOLDER,
#                        environment_folder_name=_ENVIRONMENT_FOLDER_NAME,
#                        osm_graphs_folder_name=_OSM_GRAPHS_FOLDER_NAME,
#                        graph_speeds_folder_name=_GRAPH_SPEEDS_FOLDER_NAME,
#                        gtfs_feeds_folder_name=_GTFS_FEEDS_FOLDER_NAME,
#                        schemas_folder_name=_SCHEMAS_FOLDER_NAME,
#                        models_folder_name=_MODELS_FOLDER_NAME,
#                        input_folder_name=_INPUT_FOLDER_NAME,
#                        output_folder_name=_OUTPUT_FOLDER_NAME,
#                        folder_separator="/"):
#
#         self._data_folder = data_folder
#
#         self._ENVIRONMENT_FOLDER_NAME = environment_folder_name
#         self._OSM_GRAPHS_FOLDER_NAME = osm_graphs_folder_name
#         self._GRAPH_SPEEDS_FOLDER = graph_speeds_folder_name
#         self._GTFS_FEEDS_FOLDER = gtfs_feeds_folder_name
#         self._SCHEMA_FOLDER_NAME = schemas_folder_name
#         self._MODELS_FOLDER_NAME = models_folder_name
#         self._INPUT_FOLDER_NAME = input_folder_name
#         self._OUTPUT_FOLDER_NAME = output_folder_name
#
#         self._sep = folder_separator
#
#     @property
#     def starling_sim_folder(self):
#         return starling_sim.__path__[0] + self._sep
#
#     @property
#     def schemas_folder(self):
#         return self.starling_sim_folder + self._SCHEMA_FOLDER_NAME + self._sep
#
#     @property
#     def data_folder(self):
#         return self._data_folder
#
#     @property
#     def environment_folder(self):
#         return self.data_folder + self._ENVIRONMENT_FOLDER_NAME + self._sep
#
#     @property
#     def osm_graphs_folder(self):
#         return self.environment_folder + self._OSM_GRAPHS_FOLDER_NAME + self._sep
#
#     @property
#     def graph_speeds_folder(self):
#         return self.environment_folder + self._GRAPH_SPEEDS_FOLDER + self._sep
#
#     @property
#     def gtfs_feeds_folder(self):
#         return self.environment_folder + self._GTFS_FEEDS_FOLDER + self._sep
#
#     @property
#     def models_folder(self):
#         return self.data_folder + self._MODELS_FOLDER_NAME + self._sep
#
#     def model_folder(self, model):
#         return self.models_folder + model + self._sep
#
#     def scenario_folder(self, model, scenario):
#         return self.model_folder(model) + scenario + self._sep
#
#     def scenario_input_folder(self, model, scenario):
#         return self.scenario_folder(model, scenario) + self._INPUT_FOLDER_NAME + self._sep
#
#     def scenario_output_folder(self, model, scenario):
#         return self.scenario_folder(model, scenario) + self._OUTPUT_FOLDER_NAME + self._sep
#
#
# starling_paths = StarlingPath()
#
# print(starling_paths.data_folder)
# print(starling_paths.starling_sim_folder)
# print(starling_paths.environment_folder)
# print(starling_paths.scenario_input_folder("SB_VS", "oui"))
# starling_paths.set_attributes(data_folder="tests/test_data/")
# print(starling_paths.data_folder)
# print(starling_paths.starling_sim_folder)
# print(starling_paths.environment_folder)
# print(starling_paths.scenario_input_folder("SB_VS", "oui"))
# exit(0)



import random
import logging
import numpy
import time

from starling_sim.basemodel.population.dict_population import DictPopulation
from starling_sim.basemodel.environment.environment import Environment
from starling_sim.basemodel.input.dynamic_input import DynamicInput
from starling_sim.basemodel.output.output_factory import OutputFactory
from starling_sim.basemodel.schedule.scheduler import Scheduler
from starling_sim.basemodel.trace.trace import trace_simulation_end
from starling_sim.utils.utils import import_gtfs_feed, display_horizontal_bar
from starling_sim.utils.constants import BASE_LEAVING_CODES
from starling_sim.utils.config import config


class SimulationModel:
    """
    Generic simulation model

    It must be extended to describe concrete systems,
    with their specific topologies and agents
    """

    #: Name of the simulation model
    name = "Unnamed model"

    #: Agent types of the model and their classes
    agent_type_class = None

    #: leaving codes of the model and their description
    leaving_codes = {}

    #: Agent types of the model and their modes
    modes = None

    #: Model environment class
    environment_class = Environment

    #: Model population class
    population_class = DictPopulation

    #: Model dynamic input class
    input_class = DynamicInput

    #: Model output factory class
    output_class = OutputFactory

    def __init__(self, scenario):
        """
        Initialisation of the simulation model with instances of its different elements

        :param scenario: SimulationScenario object
        """

        # simulation parameters
        self.scenario = scenario

        # random seed for the simulation setup and run
        self.randomSeed = scenario["seed"]

        # information to be completed for the specific models

        # elements  of the model
        self.environment = None
        self.agentPopulation = None
        self.dynamicInput = None
        self.outputFactory = None

        # event manager
        self.scheduler = None

        # complete, unfiltered gtfs timetable, if relevant
        self.gtfs = None

        self._base_init(scenario)

    def _base_init(self, scenario):
        self._init_model(scenario)
        self._init_scenario_run(scenario)

    def _init_model(self, scenario):
        """
        Initialise the simulation model with element that do not depend on the scenario.

        This method is called once, during the model initialisation.

        :param scenario: SimulationScenario object
        """

        # add the base leaving codes
        self.add_base_leaving_codes()

        # create the simulation environment
        self._init_environment(scenario)

    def _init_scenario_run(self, scenario):
        """
        Initialise the scenario-related elements of the model.

        This method should be called before running a new simulation scenario.

        :param scenario: SimulationScenario object
        """
        # base scenario information
        self.scenario = scenario
        self.randomSeed = scenario["seed"]

        # event manager
        self._init_scheduler()

        # model elements
        self._init_agent_population()
        self._init_dynamic_input()

        # maybe we could move the output factory to the model elements ?
        self._init_output_factory()

    def _init_environment(self, scenario):
        """
        Initialise the simulation environment using the environment_class attribute.

        :param scenario: SimulationScenario object
        """
        self.environment = self.environment_class.__new__(self.environment_class)
        self.environment.__init__(scenario)

    def _init_agent_population(self):
        """
        Initialise the simulation agent population using the population_class attribute.
        """
        self.agentPopulation = self.population_class.__new__(self.population_class)
        self.agentPopulation.__init__(self.agent_type_class.keys())

    def _init_dynamic_input(self):
        """
        Initialise the simulation dynamic input using the input_class attribute.
        """
        self.dynamicInput = self.input_class.__new__(self.input_class)
        self.dynamicInput.__init__(self.agent_type_class)

    def _init_output_factory(self):
        """
        Initialise the simulation output factory using the output_class attribute.
        """
        self.outputFactory = self.output_class.__new__(self.output_class)
        self.outputFactory.__init__()

    def _init_scheduler(self):
        """
        Initialise the simulation scheduler.
        """
        self.scheduler = Scheduler()

    def setup(self):
        """
        Sets the entries of the simulation model and prepares it for the simulation run.
        It is here that the files and data are effectively read and imported in the model.

        This method can be extended to manage and setup other elements of the model
        """
        self._setup_model()
        self._setup_scenario_run()

    def _setup_model(self):
        """
        Set up the simulation model elements that do not depend on the scenario.

        This method is called once, during the model setup.
        """
        start = time.time()

        logging.info("Setting up model")
        # setup model elements
        self._setup_environment()
        self.setup_gtfs()

        duration = round(time.time() - start, 1)
        logging.info("End of model setup. Elapsed time : {} seconds\n".format(duration))
        self.scenario.set_stat("model_setup_time", duration)

    def _setup_scenario_run(self):
        """
        Set up the scenario-related elements of the model.

        This method should be called before running a new simulation scenario.
        """

        display_horizontal_bar()
        logging.info("Setting up run of scenario: {}".format(self.scenario.name))
        start = time.time()
        # set the parameters and initialize the random seed
        self.setup_seeds()

        # setup scenario elements
        self._setup_dynamic_input()
        self._setup_output_factory()

        duration = round(time.time() - start, 1)
        logging.info("End of scenario setup. Elapsed time : {} seconds\n".format(duration))
        self.scenario.set_stat("scenario_setup_time", duration)

    def _setup_environment(self):
        """
        Set up the simulation environment.
        """
        logging.debug("Simulation environment setup")
        self.environment.setup(self)

    def _setup_dynamic_input(self):
        """
        Set up the simulation dynamic input.
        """
        logging.debug("Dynamic input setup")
        self.dynamicInput.setup(self)

    def _setup_output_factory(self):
        """
        Set up the simulation output factory.
        """
        logging.debug("Output factory setup")
        self.outputFactory.setup(self)

    def run(self):
        """
        Runs the simulation model until the time limit is reached.

        This method can be extended to run other elements of the model
        """

        logging.info("Starting simulation run")

        start = time.time()

        # if asked, add a process that logs the simulation time every hour
        if "time_log" in self.scenario and self.scenario["time_log"]:
            self.scheduler.new_process(self.periodic_hour_log())

        # create agents and add their loops
        self.scheduler.new_process(self.dynamicInput.play_dynamic_input_())

        # run the simulation
        self.scheduler.run(self.scenario["limit"])

        # trace the end of the simulation for all agents
        trace_simulation_end(self)

        duration = round(time.time() - start, 1)

        logging.info("End of simulation run. Elapsed time : {} seconds\n".format(duration))
        self.scenario.set_stat("execution_time", duration)

        shortest_path_count = 0
        for topology in self.environment.topologies.values():
            shortest_path_count += topology.shortest_path_count
        self.scenario.set_stat("shortest_paths", shortest_path_count)

    def generate_output(self):
        """
        Generate an output of the current simulation
        """

        logging.info("Generating outputs")
        self.outputFactory.extract_simulation(self)

    def add_base_leaving_codes(self):
        """
        Add the base leaving codes to the ones specified for the model.

        This will overwrite any custom code named as a base code.
        """

        self.leaving_codes.update(BASE_LEAVING_CODES)

    def setup_seeds(self):
        """
        Set the seeds of the random functions
        """

        random.seed(self.randomSeed)
        numpy.random.seed(self.randomSeed)

    def periodic_hour_log(self):
        while True:
            logging.info("Current simulation time is {}".format(self.scheduler.now()))

            yield self.scheduler.timeout(3600)

    def setup_gtfs(self):
        """
        Set up a gtfs timetable for the simulation.
        """

        if "gtfs_timetables" in self.scenario:
            logging.debug("GTFS tables setup")
            # import the gtfs timetable from the zip given in the parameters
            restrict_transfers = config["transfer_restriction"]
            self.gtfs = import_gtfs_feed(self.scenario["gtfs_timetables"], restrict_transfers)

    @classmethod
    def get_agent_type_schemas(cls):
        agent_type_class = cls.agent_type_class
        schemas = dict()
        for agent_type in agent_type_class.keys():
            schemas[agent_type] = agent_type_class[agent_type].get_schema()

        return schemas



from starling_sim.basemodel.topology.network_weight import NetworkWeight
import random


class BikeWeightOSM(NetworkWeight):
    """
    Improve bike routing by defining a better weight on bike networks.

    This weight is defined on OSM attributes. The necessary attributes are:
        - 'cycleway', 'cycleway:right', 'cycleway:left'
        - 'bridge'

    """

    DEFAULT_PARAMETERS = {
        "length": 2.25,
        "link_constant": 1.61,
        "upslope>4.length": 3.24,
        "facility.length": -0.74,
        "separate_facility.length": -1.80,
        "bridge": 5.41,
        "bridge.facility": -2.83,
    }

    def pre_process_edge(self, u, v, d):
        # d["slope"] = random.randint(-8, 8)*random.random()

        cycleway = None

        if "highway" in d and d["highway"] == "cycleway":
            cycleway = "separate"

        elif "cycleway" in d:
            cycleway = d["cycleway"]
        elif "cycleway:left" in d:
            cycleway = d["cycleway:left"]
        elif "cycleway:right" in d:
            cycleway = d["cycleway:right"]

        if cycleway == "no":
            bike_facility = "no"
        elif cycleway == "track" or cycleway == "separate":
            bike_facility = "separate"
        else:
            bike_facility = "yes"

        d["bike_facility"] = bike_facility

    def compute_edge_weight(self, u, v, d, parameters):
        total_weight = 0
        length = d["length"]

        total_weight += parameters["link_constant"]

        total_weight += parameters["length"] * length

        # if d["slope"] > 4:
        #     total_weight += parameters["upslope>4.length"]*length

        if d["bike_facility"] == "separate":
            total_weight += parameters["separate_facility.length"] * length
        elif d["bike_facility"] == "yes":
            total_weight += parameters["facility.length"] * length

        if "bridge" in d and d["bridge"] == "yes":
            total_weight += parameters["bridge"]

            if d["bike_facility"] in ["separate", "yes"]:
                total_weight += parameters["bridge.facility"]

        if total_weight < 0:
            total_weight = 0.01

        return total_weight



from starling_sim.basemodel.agent.vehicles.vehicle import Vehicle


class StationBasedVehicle(Vehicle):
    """
    This class describes a vehicle of a station-based vehicle sharing service.
    """

    SCHEMA = {
        "properties": {
            "station": {
                "title": "Origin station",
                "description": "Origin station of the vehicle",
                "type": "string",
            }
        },
        "required": ["station"],
        "remove_props": ["origin"],
    }

    def __init__(self, simulation_model, agent_id, seats, station, **kwargs):
        """
        Get information from the origin station and update its stock.

        :param simulation_model:
        :param agent_id:
        :param seats:
        :param kwargs:
        """

        # get the station agent
        station_agent = simulation_model.agentPopulation.get_agent(station)
        if station_agent is None:
            raise ValueError(
                "Could not match the station id '{}' with any simulation "
                "agent (initialising vehicle  {})".format(station, agent_id)
            )

        # update the station's store
        station_agent.store.items.append(self)
        station_agent.initial_stock += 1

        super().__init__(simulation_model, agent_id, station_agent.origin, seats, **kwargs)



import logging
import osmnx as ox
import networkx as nx
import numpy as np

from starling_sim.basemodel.topology.topology import Topology


class EmptyNetwork(Topology):
    """
    An initially empty topology
    """

    def __init__(self, transport_mode, weight_class=None, store_paths=False):
        """
        Create the topology structure, without initializing the network

        :param transport_mode: type of the given network
        :param weight_class: class used for defining weight
        :param store_paths: boolean indicating if shortest paths should be stored
        """
        super().__init__(transport_mode, weight_class=weight_class, store_paths=store_paths)

        self.graph = None
        self.speeds = None

    def init_graph(self):
        # initialise an empty graph object
        logging.debug("Generating an empty graph for mode '{}'".format(self.mode))
        self.graph = nx.MultiDiGraph()

    def add_time_and_length(self, u, v, d):
        pass

    def position_localisation(self, position):
        return [self.graph.nodes[position]["y"], self.graph.nodes[position]["x"]]

    def nearest_position(self, localisation):
        return ox.distance.nearest_nodes(self.graph, float(localisation[1]), float(localisation[0]))

    def localisations_nearest_nodes(self, x_coordinates, y_coordinates, return_dist=False):
        # convert coordinate lists to np.array
        x_array = np.array(x_coordinates, dtype=np.float32)
        y_array = np.array(y_coordinates, dtype=np.float32)

        # we use osmnx nearest_nodes function
        return ox.distance.nearest_nodes(self.graph, x_array, y_array, return_dist=return_dist)



from starling_sim.basemodel.topology.network_weight import NetworkWeight


class SimpleTimeWeight(NetworkWeight):
    """
    Simply define the weight as equal to the time spent on the edge.

    This results in shortest (time) paths.
    """

    def compute_edge_weight(self, u, v, d, parameters):
        return d[self.topology.TIME_ATTRIBUTE]



from starling_sim.basemodel.simulation_model import SimulationModel
from starling_sim.models.FF_VS.output import Output
from starling_sim.models.FF_VS.user import User
from starling_sim.basemodel.agent.vehicles.vehicle import Vehicle


class Model(SimulationModel):
    """
    Simple model for the simulation of a free-floating vehicle-sharing system
    """

    name = "Free-floating vehicle-sharing model"

    agent_type_class = {"user": User, "vehicle": Vehicle}

    leaving_codes = {"FAIL_GET": "Failed to get a vehicle"}

    modes = {"user": ["walk"], "vehicle": [None, "walk"]}

    output_class = Output



import copy

from starling_sim.basemodel.trace.trace import Traced
from starling_sim.basemodel.trace.events import InputEvent
from starling_sim.basemodel.agent.operators.operator import Operator
from starling_sim.utils.utils import json_load, validate_against_schema, add_defaults_and_validate
from starling_sim.utils.constants import STOP_POINT_POPULATION
from starling_sim.utils.paths import scenario_agent_input_filepath
from jsonschema import ValidationError
from json import JSONDecodeError

import traceback
import random
import os
from copy import deepcopy


class DynamicInput(Traced):
    """
    This class manages the agent entering the simulation.

    It generates base agents at the beginning of the simulation, and
    dynamically adds agents in the environment during the simulation.
    """

    DUPLICATE_AGENT_ID_FORMAT = "{original_id}.{index}"

    def __init__(self, agent_type_dict):
        super().__init__("INPUT")

        #: correspondence dict between agent_type and class
        self.agent_type_class = agent_type_dict

        self.agent_type_schemas = None

        self.dynamic_feature_list = None

    def __str__(self):
        """
        Gives a string display to the dynamic input
        :return:
        """

        return "[dynamicInput: model={}, randomSeed={}]".format(self.sim.name, self.sim.randomSeed)

    def setup(self, simulation_model):
        """
        Set the simulation model attribute and generate the base agents.

        :param simulation_model: SimulationModel
        """

        # set the simulation model attribute
        self.sim = simulation_model

        self.agent_type_schemas = self.sim.get_agent_type_schemas()

        # set the attribute of dynamic features

        self.dynamic_feature_list = self.feature_list_from_file(
            self.sim.scenario["dynamic_input_file"]
        )

        # sort list according to origin times
        self.dynamic_feature_list = sorted(
            self.dynamic_feature_list, key=lambda x: x["properties"]["origin_time"]
        )

        # get the list of static features (present at the start of the simulation)
        init_files = self.sim.scenario["init_input_file"]

        # if there are several files, concatenate their feature lists
        if isinstance(init_files, list):
            init_feature_list = []
            for filename in init_files:
                init_feature_list += self.feature_list_from_file(filename)
        else:
            init_feature_list = self.feature_list_from_file(init_files)

        # resolve the modes of the agent types
        self.resolve_type_modes_from_inputs(init_feature_list + self.dynamic_feature_list)

        init_without_operators = []

        # create the operators agents
        for feature in init_feature_list:
            agent_type = feature["properties"]["agent_type"]
            agent_class = self.agent_type_class[agent_type]

            if issubclass(agent_class, Operator):
                self.new_agent_input(feature)
            else:
                init_without_operators.append(feature)

        # pre process the positions of the rest of the init input
        self.pre_process_position_coordinates(init_without_operators)

        # create the rest of the init input
        for feature in init_without_operators:
            # generate a new agent based on the feature properties
            self.new_agent_input(feature)

        # pre process the positions of the dynamic input
        self.pre_process_position_coordinates(self.dynamic_feature_list)

    def feature_schema_validation(self, feature):
        # validate against Feature schema
        validate_against_schema(feature, "geojson/Feature.json")

        # test if the feature has an 'agent_type' property
        if "agent_type" not in feature["properties"]:
            raise KeyError(
                "Error in {} : input features must contain an 'agent_type' property".format(feature)
            )

        # validate and set defaults using the schema corresponding to the agent type
        agent_schema = self.agent_type_schemas[feature["properties"]["agent_type"]]
        props = feature["properties"]
        final_props = add_defaults_and_validate(props, agent_schema)
        feature["properties"] = final_props

        return feature

    def play_dynamic_input_(self):
        """
        Add agents to the simulation over time.

        Agents are created based on the dynamic input file,
        where the generation time is specified as 'origin_time'.
        """

        for feature in self.dynamic_feature_list:
            # TODO : check the feature schema ? duplicate with FeatureCollection check

            # see if an offset should be applied to the input origin time
            if (
                "early_dynamic_input" in self.sim.scenario
                and self.sim.scenario["early_dynamic_input"]
            ):
                early_input_time_offset = self.sim.scenario["early_dynamic_input"]
            else:
                early_input_time_offset = 0

            # compute the effective generation time
            generation_time = int(feature["properties"]["origin_time"]) - early_input_time_offset

            # check that the generation time is positive
            if generation_time < 0:
                self.log_message(
                    "Feature {} cannot be generated at {}, "
                    "generation at time 0 instead.".format(feature, generation_time),
                    30,
                )
                generation_time = 0

            # wait for the next generation
            waiting_time = generation_time - self.sim.scheduler.now()
            yield self.sim.scheduler.timeout(waiting_time)

            # generate new agent
            self.new_agent_input(feature)

    def new_agent_input(self, feature):
        """
        Create and initialise a new agent, and add it to the simulation environment.

        :param feature: agent feature (GeoJSON Feature)

        :return: created agent or list of agents in case of "duplicates"
        """

        # get the agent input dict
        input_dict = feature["properties"]

        # if 'duplicates', run new_agent_input recursively on duplicated features and return a list of new agents
        if "duplicates" in input_dict:
            new_agents = []
            nb_duplicates = input_dict["duplicates"]
            # check agent properties
            assert (
                isinstance(nb_duplicates, int) and nb_duplicates > 0
            ), "'duplicates' attribute must be a strictly positive integer"
            assert "agent_id" in input_dict, "Duplicated agent misses agent_id attribute"

            # create a new agent feature for each duplicate
            for i in range(nb_duplicates):
                duplicated_feature = copy.deepcopy(feature)
                del duplicated_feature["properties"]["duplicates"]
                # create a new id using the original agent id and the duplicate index
                duplicated_feature["properties"]["agent_id"] = (
                    self.DUPLICATE_AGENT_ID_FORMAT.format(
                        original_id=input_dict["agent_id"], index=i
                    )
                )

                new_agents.append(self.new_agent_input(duplicated_feature))

            # return a list of agents instead of a single object
            return new_agents

        if "operator_id" in input_dict:
            # link with operator
            self.add_key_operator(input_dict)

        # pre-process the agent input dict
        self.pre_process_input_dict(input_dict)

        # validate the feature and add default values
        try:
            feature = self.feature_schema_validation(feature)
        except Exception as e:
            self.log_message(
                "Agent input was not completed due to the following error : {}".format(str(e)), 30
            )
            return

        input_dict = feature["properties"]

        # the agent is associated to the population of its type and other provided populations
        if "population" in input_dict:
            populations = input_dict["population"]
            if isinstance(populations, list):
                populations.append(input_dict["agent_type"])
            else:
                populations = [populations, input_dict["agent_type"]]

            # only keep distinct populations
            populations = set(populations)
            populations = list(populations)
        else:
            populations = input_dict["agent_type"]

        # get the agent type
        agent_type = input_dict["agent_type"]

        if agent_type in self.agent_type_class:
            # get the class to generate
            agent_class = self.agent_type_class[agent_type]

            # generate the new agent
            new_agent = agent_class.__new__(agent_class)

            # initialise the new agent
            try:
                new_agent.__init__(self.sim, **input_dict)
            except (TypeError, KeyError, ValidationError):
                # if the initialisation fails, log and leave
                self.log_message(
                    "Instantiation of {}  failed with message :\n {}".format(
                        self.agent_type_class[agent_type], traceback.format_exc()
                    ),
                    30,
                )
                exit(1)

            # add the agent to the simulation environment
            self.add_agent_to_simulation(new_agent, populations)

            return new_agent

        else:
            self.log_message(
                "Unknown agent_type {}. Model agent types are {}.".format(
                    agent_type, list(self.agent_type_class.keys())
                ),
                30,
            )
            return

    def add_agent_to_simulation(self, agent, populations):
        """
        Add the agent to the simulation environment.

        Add the agent to its population, then trace an input event
        and start its simpy loop.

        :param agent: Agent object
        :param populations: population(s) where the agent belongs
        """

        # cancel add to simulation if agent already exists
        try:
            # add agent to relevant population
            self.sim.agentPopulation.new_agent_in(agent, populations)

            # trace and log input event
            self.trace_event(InputEvent(self.sim.scheduler.now(), agent))

            # add the agent loop to the event manager
            agent.main_process = self.sim.scheduler.new_process(agent.simpy_loop_())
        except KeyError:
            pass

    # get and manage input dicts from the input files

    def feature_list_from_file(self, filename):
        """
        Get the list of input features from the given filename.

        The file must be a geojson, following the FeatureCollection schema,
        and be stored in the input folder.

        :param filename: name of the input file, stored in the input folder
        :return: list of geojson Feature dicts
        """

        if filename is None:
            return []

        # get the path to the input file
        filepath = scenario_agent_input_filepath(self.sim.scenario.scenario_folder, filename)

        # read the dict contained in input file
        try:
            geojson_input = json_load(filepath)

            # TODO : validate against FeatureCollection

        except JSONDecodeError as e:
            self.log_message(
                "Error while decoding input file {} : {}\n "
                "Are you sure the file is a JSON ?".format(filename, e),
                40,
            )
            raise e

        except ValidationError as e:
            self.log_message(
                "Error while validating the input data : {}\n Are you sure "
                "the json follows the FeatureCollection schema ?".format(e),
                40,
            )
            raise e

        # return the feature list
        return geojson_input["features"]

    def make_demand_static(self):
        if "make_static" in self.sim.scenario and self.sim.scenario["make_static"] in [
            "all",
            "prebooked",
            "prebooked_only",
            "ghosts",
        ]:
            # also add the agents of dynamic input that are prebooked
            dynamic_features = []

            make_static = self.sim.scenario["make_static"]

            for feature in self.dynamic_feature_list:
                properties = feature["properties"]

                if make_static == "all":
                    self.new_agent_input(feature)
                elif make_static == "prebooked" and properties["prebooked"]:
                    self.new_agent_input(feature)
                elif make_static == "prebooked_only":
                    if properties["prebooked"]:
                        self.new_agent_input(feature)
                elif make_static == "ghosts":
                    self.new_agent_input(feature)
                else:
                    dynamic_features.append(feature)

            # store the dynamic ones back
            self.dynamic_feature_list = dynamic_features

    def pre_process_position_coordinates(self, features):
        """
        Add a position to the features with coordinates inputs.

        Group the features by modes and call localisations_nearest_nodes environment method,
        then update the features with the resulting positions.

        :param features: features to pre process
        """

        # create a global dict that associates modes to a list of features and their information
        pre_process_dict = dict()

        # get the model modes dict
        model_modes = self.sim.modes

        # base structure of the content associated to modes
        base_nearest_nodes_dict = {
            "inputs": [],
            "keys": [],
            "lon": [],
            "lat": [],
            "nearest_nodes": None,
        }

        # browse the features
        for feature in features:
            input_dict = feature["properties"]

            # prepare structures for storing feature information
            inputs = []
            keys = []
            lon = []
            lat = []

            # look for coordinates inputs

            origin_coordinates = self.get_position_coordinates_from_feature(feature, "origin")
            if origin_coordinates is not None:
                # deprecated properties 'origin_lon' and 'origin_lat'
                if (
                    origin_coordinates == [0, 0]
                    and "origin_lon" in input_dict
                    and "origin_lat" in input_dict
                ):
                    self.log_message(
                        "Use of 'origin_lon' and 'origin_lat' is deprecated, "
                        "using the feature geometry is preferred",
                        30,
                    )
                    origin_coordinates = [input_dict["origin_lon"], input_dict["origin_lat"]]

                if origin_coordinates != [0, 0]:
                    inputs.append(input_dict)
                    keys.append("origin")
                    lon.append(origin_coordinates[0])
                    lat.append(origin_coordinates[1])

            destination_coordinates = self.get_position_coordinates_from_feature(
                feature, "destination"
            )
            if destination_coordinates is not None:
                # deprecated properties 'destination_lon' and 'destination_lat'
                if (
                    destination_coordinates == [0, 0]
                    and "destination_lon" in input_dict
                    and "destination_lat" in input_dict
                ):
                    self.log_message(
                        "Use of 'destination_lon' and 'destination_lat' is deprecated, "
                        "using the feature geometry is preferred",
                        30,
                    )
                    destination_coordinates = [
                        input_dict["destination_lon"],
                        input_dict["destination_lat"],
                    ]

                if destination_coordinates != [0, 0]:
                    inputs.append(input_dict)
                    keys.append("destination")
                    lon.append(destination_coordinates[0])
                    lat.append(destination_coordinates[1])

            # if there are coordinates inputs, add them to the modes dict
            if len(inputs) != 0:
                # get the modes of the input
                modes = model_modes[input_dict["agent_type"]]

                # if the dict does not exist, create it
                if modes not in pre_process_dict:
                    pre_process_dict[modes] = deepcopy(base_nearest_nodes_dict)

                # append the information to the dict
                nearest_nodes_dict = pre_process_dict[modes]
                nearest_nodes_dict["inputs"] += inputs
                nearest_nodes_dict["keys"] += keys
                nearest_nodes_dict["lon"] += lon
                nearest_nodes_dict["lat"] += lat

        # for each mode group, compute the localisations
        for modes in pre_process_dict.keys():
            # call localisations_nearest_nodes on the dict information
            nearest_nodes_dict = pre_process_dict[modes]
            nearest_nodes = self.sim.environment.localisations_nearest_nodes(
                nearest_nodes_dict["lon"], nearest_nodes_dict["lat"], list(modes)
            )
            nearest_nodes_dict["nearest_nodes"] = nearest_nodes

            # affect the positions back to the input dicts
            for i in range(len(nearest_nodes_dict["inputs"])):
                input_dict = nearest_nodes_dict["inputs"][i]
                input_dict[nearest_nodes_dict["keys"][i]] = nearest_nodes_dict["nearest_nodes"][i]

    def get_position_coordinates_from_feature(self, feature, position_key):
        geometry_type = feature["geometry"]["type"]
        geometry_coordinates = feature["geometry"]["coordinates"]

        res_coordinates = None

        if position_key == "origin":
            if geometry_type == "Point":
                res_coordinates = geometry_coordinates
            elif geometry_type == "LineString":
                res_coordinates = geometry_coordinates[0]

        elif position_key == "destination":
            if geometry_type == "LineString":
                res_coordinates = geometry_coordinates[-1]

        else:
            self.log_message("Unsupported position key '{}'".format(position_key))

        return res_coordinates

    def resolve_type_modes_from_inputs(self, features):
        """
        Resolve the model modes from the inputs.

        Browse the inputs and resolve the missing values of the modes dict.
        Raise an error if there are conflicting values for a same mode.

        :param features: list of input features
        :raises ValueError: if there are problem during the mode resolve
        """

        # get the modes dict of the model
        model_modes = self.sim.modes
        if model_modes is None:
            raise ValueError("Model modes are not specified")

        # do a first pass without replacing the type references
        # only resolve None values and check for conflicts
        for feature in features:
            # get the input dict and its associated mode values
            input_dict = feature["properties"]
            type_modes = model_modes[input_dict["agent_type"]]

            # get an eventual input value
            if "mode" in input_dict:
                input_value = input_dict["mode"]
            else:
                input_value = None

            mode = None

            if isinstance(type_modes, list):
                for i in range(len(type_modes)):
                    mode = self.resolve_mode(type_modes, i, input_value, False)

            if isinstance(type_modes, dict):
                for key in type_modes.keys():
                    # get the relevant input value
                    if input_value is not None:
                        val = input_value[key]
                    else:
                        val = None

                    mode = self.resolve_mode(type_modes, key, val, False)

            # affect the resulting mode if no mode was specified
            if input_value is None and mode is not None:
                input_dict["mode"] = mode

        # do a second pass to resolve the type references
        for agent_type in model_modes.keys():
            modes = model_modes[agent_type]

            if isinstance(modes, list):
                for i in range(len(modes)):
                    self.resolve_mode(modes, i, None, True)

                # transform the lists into sorted tuples without duplicates
                model_modes[agent_type] = tuple(sorted(set(modes)))

            if isinstance(modes, dict):
                for key in modes.keys():
                    self.resolve_mode(modes, key, None, True)

    def resolve_mode(self, obj, key, input_value, replace_types):
        """
        Resolve the object mode value with recursive calls.

        Resolution is done depending on the nature of the mode value:

        - if the mode value is a topology, return it
        - if the mode value is an agent type, resolve the agent type first mode value
        - if the mode value is None, return the input value

        Input values are only used for the first mode value of lists or the mode values of dicts.

        Check the final result against the input value if provided.

        :param obj: object containing the mode values
        :param key: key of the resolved mode value
        :param input_value: input value or None
        :param replace_types: boolean indicating if agent type value should be
            replaced with the resolved mode.

        :return: resolved mode or None if not resolved
        """

        # only consider input value for first list values or dict values
        if isinstance(key, str) or key == 0:
            input_value = input_value
        else:
            input_value = None

        # get the list of topologies and the global dict of modes
        topologies = list(self.sim.environment.topologies.keys())
        agent_type_modes = self.sim.modes

        # get the keyword to replace
        keyword = obj[key]

        mode = None

        # if the mode value is a topology, return it
        if keyword in topologies:
            mode = keyword

        # if the mode value is an agent type, resolve the agent type first mode value
        elif keyword in agent_type_modes:
            # resolve the agent type mode first mode value
            mode = self.resolve_mode(agent_type_modes[keyword], 0, input_value, replace_types)

            # replace the agent type value with mode if asked
            if replace_types:
                obj[key] = mode

        # if the mode value is None, return the input value
        elif keyword is None:
            obj[key] = input_value
            mode = input_value

        # check the resulting mode with input value if mode value or dict object
        if input_value is not None and mode != input_value:
            raise ValueError(
                "Conflict between type mode '{}' and input value '{}'".format(mode, input_value)
            )

        return mode

    def pre_process_input_dict(self, input_dict):
        """
        Enhance the given input dict according to the agent type and the needs of the model.

        :param input_dict: input dict to be completed
        :return:
        """
        pass

    def add_key_position_from_stop_point(self, input_dict, key):
        input_dict_key = key + "_stop_point"

        # if an operator is provided, use its stop points
        if "operator_id" in input_dict:
            stop_points_dict = self.sim.agentPopulation.get_agent(
                input_dict["operator_id"]
            ).stopPoints
        # otherwise fetch them from the global stop points population
        else:
            stop_points_dict = self.sim.agentPopulation[STOP_POINT_POPULATION]

        if input_dict[input_dict_key] == "random":
            stop_point = random.choice(list(stop_points_dict.values()))
        else:
            stop_point = stop_points_dict[input_dict[input_dict_key]]

        input_dict[key] = stop_point.position

    def add_key_operator(self, input_dict):
        # get the operator id, look in the input dict if not provided
        operator_id = input_dict["operator_id"]

        # get the operator
        operator = self.sim.agentPopulation.get_agent(operator_id)

        # get position from eventual depot
        if "depot" in input_dict:
            depot = operator.depotPoints[input_dict["depot"]]
            input_dict["origin"] = depot.position

        populations = [input_dict["agent_type"]]
        if "population" in input_dict:
            populations.append(input_dict["population"])

        if operator.fleet_name in populations:
            input_service = "fleet"
        elif operator.staff_dict_name in populations:
            input_service = "staff"
        else:
            return

        # set others keys according to input service
        if input_service is None:
            return
        elif input_service in ["fleet", "staff"]:
            # set the operator fleet mode if not already done
            # SUPPOSITION : all fleet and staff use the same mode
            if operator.mode is None:
                operator.mode = {input_service: input_dict["mode"]}
            elif input_service not in operator.mode or operator.mode[input_service] is None:
                operator.mode[input_service] = input_dict["mode"]
            else:
                # set input mode from operator if no mode provided
                if "mode" not in input_dict or input_dict["mode"] is None:
                    input_dict["mode"] = operator.mode[input_service]

                # if both are provided and are different, log an error and exit
                elif operator.mode[input_service] != input_dict["mode"]:
                    self.log_message(
                        "Input mode of '{}' agent {} ({}) differs from operator's mode ({})".format(
                            input_service,
                            input_dict["agent_id"],
                            input_dict["mode"],
                            operator.mode[input_service],
                        ),
                        40,
                    )
                    raise ValueError("Conflict between operator's mode and its fleet/staff's mode")

            # set the input population
            if input_service == "fleet":
                input_dict["population"] = operator.fleet_name
            else:
                input_dict["population"] = operator.staff_dict_name
        else:
            self.log_message("Unknown input service type {}".format(input_service))



from starling_sim.basemodel.agent.spatial_agent import SpatialAgent


class Station(SpatialAgent):
    """
    Station agent
    """

    SCHEMA = {
        "properties": {
            "operator_id": {
                "title": "Operator ID",
                "description": "ID of the operator managing this agent",
                "type": "string",
            }
        }
    }

    def __init__(self, simulation_model, agent_id, origin, operator_id=None, **kwargs):
        SpatialAgent.__init__(self, simulation_model, agent_id, origin, **kwargs)

        self.operator = None
        if operator_id is not None:
            self.operator = self.sim.agentPopulation.get_agent(operator_id)

    def __str__(self):
        return "[id={}, position={}]".format(self.id, self.position)



import logging
import osmnx as ox
import numpy as np

from starling_sim.basemodel.topology.topology import Topology
from starling_sim.utils.paths import graph_speeds_folder
from starling_sim.utils.utils import json_load, osm_graph_from_file


class OSMNetwork(Topology):
    """
    Open Street Map network topology
    """

    def __init__(
        self,
        transport_mode,
        weight_class=None,
        store_paths=False,
        network_file=None,
        speed_file=None,
        graph=None,
    ):
        """
        Create the topology structure, without initializing the network

        :param transport_mode: type of the given network
        :param weight_class: class used for defining weight
        :param store_paths: boolean indicating if shortest paths should be stored
        :param network_file: graphml file for network import
        :param speed_file: json file containing the link speeds
        :param graph: graph object for direct initialisation
        """
        super().__init__(transport_mode, weight_class=weight_class, store_paths=store_paths)

        self.network_file = network_file
        self.speed_file = speed_file

        self.graph = graph
        self.speeds = None

    def init_graph(self):
        if self.graph is None:
            if self.network_file is None:
                logging.error("No network file provided for topology initialisation")
                raise ValueError("Network file provided is {}".format(self.network_file))
            else:
                logging.debug(
                    "Importing OSM graph for mode '{}' from file {}".format(
                        self.mode, self.network_file
                    )
                )
                self.graph = osm_graph_from_file(self.network_file)

        if self.speed_file is None:
            logging.error("No speed file provided for topology initialisation")
            raise ValueError("Speed file provided is {}".format(self.network_file))
        else:
            logging.debug(
                "Importing graph speeds for mode '{}' from file {}".format(
                    self.mode, self.speed_file
                )
            )
            self.speeds = json_load(graph_speeds_folder() + self.speed_file)

    def add_time_and_length(self, u, v, d):
        """
        Set the edge time and length using the speeds dict and OSM length values.

        Speeds are set according to OSM 'highway' attribute.
        If the attribute's value doesn't correspond to any key, use the 'other' key.

        :param u:
        :param v:
        :param d:
        """

        if isinstance(d["highway"], list):
            d["highway"] = d["highway"][0]

        # differentiate speeds among the link types
        if d["highway"] in self.speeds:
            speed = self.speeds[d["highway"]]["speed"]
        elif d["highway"].endswith("_link") and d["highway"][:-5] in self.speeds:
            speed = self.speeds[d["highway"][:-5]]["speed"]
        else:
            speed = self.speeds["other"]["speed"]

        d["speed"] = speed

        # round the length values
        d[self.LENGTH_ATTRIBUTE] = round(d[self.LENGTH_ATTRIBUTE])

        # time is valued in seconds
        d[self.TIME_ATTRIBUTE] = round(3600 * ((d[self.LENGTH_ATTRIBUTE] / 1000) / speed))

    def position_localisation(self, position):
        return [self.graph.nodes[position]["y"], self.graph.nodes[position]["x"]]

    def nearest_position(self, localisation):
        return ox.distance.nearest_nodes(self.graph, float(localisation[1]), float(localisation[0]))

    def localisations_nearest_nodes(self, x_coordinates, y_coordinates, return_dist=False):
        # convert coordinate lists to np.array
        x_array = np.array(x_coordinates, dtype=np.float32)
        y_array = np.array(y_coordinates, dtype=np.float32)

        # we use osmnx nearest_nodes function
        return ox.distance.nearest_nodes(self.graph, x_array, y_array, return_dist=return_dist)



"""
This model describes a station-based vehicle-sharing system.
This package contains the classes that extend the basemodel and implement the model's features.
More information on the :ref:`SB_VS` page.
"""



from starling_sim.basemodel.output.kpis import KPI

import logging
import pandas as pd
import os
import math
from datetime import datetime, time, timedelta


KEY_TIME_RANGE = "timeRange"


class KpiOutput:
    def __init__(self, population_names, kpi_list, kpi_name=None):
        # simulation model access
        self.sim = None

        # name of the kpi, will compose the kpi filename : <kpi_name>.csv
        if kpi_name is None:
            if isinstance(population_names, list):
                self.name = "_&_".join(population_names) + "_kpi"
            else:
                self.name = population_names + "_kpi"
        else:
            self.name = kpi_name

        # population of agent to evaluate
        self.population_names = population_names
        self.populations = None

        # list of kpi to evaluate the given agents
        self.kpi_list = kpi_list
        self.columns = None

        # dict containing kpi values
        self.kpi_rows = None

        # indicates if KpiOutput is time profiled
        self.time_profiling = None

        # output file
        self.filename = None
        self.folder = None

    def setup(self, filename, folder, simulation_model):
        """
        Setup method called during simulation setup.

        Sets the values of out file and folder, and call setup for KPIs.

        :param filename: .csv file
        :param folder:
        :param simulation_model:
        :return:
        """

        self.sim = simulation_model
        self.filename = filename
        self.folder = folder

        # setup kpis and get columns
        columns = [KPI.KEY_ID]
        time_profiling = None
        for kpi in self.kpi_list:
            # setup kpi
            kpi.setup(self, simulation_model)
            if time_profiling is None:
                time_profiling = kpi.profile is not None
            else:
                assert time_profiling == (
                    kpi.profile is not None
                ), f"KPIs profiling types cannot be mixed ({self.name})"

            # add kpi columns
            columns += kpi.export_keys

        self.time_profiling = time_profiling
        if self.time_profiling:
            columns.insert(1, KEY_TIME_RANGE)
        self.columns = columns

        if isinstance(self.population_names, list):
            self.populations = [
                simulation_model.agentPopulation[population_name]
                for population_name in self.population_names
            ]
        else:
            self.populations = [simulation_model.agentPopulation[self.population_names]]

    def write_kpi_table(self):
        """
        Write the KPI of the population in the csv file
        obtained from out file attributes
        The KPIs evaluated are defined by the kpi_list attribute
        """

        # build the KPI table for all agents of each population
        kpi_table = self.build_kpi_table()

        if self.time_profiling:
            kpi_table[KEY_TIME_RANGE] = kpi_table[KEY_TIME_RANGE].apply(
                lambda x: (datetime.min + timedelta(seconds=x)).strftime("%H:%M:%S")
            )

        # do not generate a kpi output if the kpi table is empty
        if kpi_table.empty:
            return

        path = str(os.path.join(self.folder, self.filename))
        try:
            # write the dataframe into a csv file
            kpi_table.to_csv(path, sep=";", index=False)

            # signal new file to output factory
            mimetype = "text/csv"
            if path.endswith("gz"):
                mimetype = "application/gzip"
            self.sim.outputFactory.new_output_file(
                path,
                mimetype,
                compressed_mimetype="text/csv",
                content="kpi",
                subject=self.name,
            )

        except KeyError as e:
            logging.warning(
                "Could not generate kpi output {}, " "error occurred : {}".format(path, e)
            )

    def build_kpi_table(self) -> pd.DataFrame:
        """
        Build the output KPI table.

        KPIs from self.kpis are evaluated on each agent of each population from self.populations.

        :return: KPI DataFrame
        """

        kpi_tables = []
        # compute kpi tables for each population dict
        for population in self.populations:
            # compute a kpi table for each agent of the population
            for agent in population.values():
                kpi_tables.append(self.compute_agent_kpis(agent))

        result = pd.concat(kpi_tables)
        return result

    def compute_agent_kpis(self, agent):
        """
        Build a DataFrame containing indicator evaluated on the given agent.

        The DataFrame columns are defined by the KPIs `keys` attributes,
        with and additional column for the agent id, and an optional column
        for time profiling.

        The DataFrame can contain several rows, for instance when KPIs
        are profiled by time.

        :param agent: Agent on which KPIs are evaluated

        :return: DataFrame containing indicators evaluated on the given agent
        """

        self.kpi_rows = {key: [] for key in self.columns}

        # evaluate indicators on agent
        for kpi in self.kpi_list:
            kpi.evaluate_for_agent(agent)

        self.kpi_rows[KPI.KEY_ID] = agent.id
        if self.time_profiling:
            self.kpi_rows[KEY_TIME_RANGE] = self.kpi_list[0].profile

        res = pd.DataFrame(self.kpi_rows, columns=self.columns)

        return res



from starling_sim.basemodel.agent.vehicles.repositioning_vehicle import RepositioningVehicle


class RepositioningStaff(RepositioningVehicle):
    """
    This class describes the repositioning staff of the SB_VS_R model.
    """

    SCHEMA = {"properties": {}, "required": ["dwell_time"]}

    def __init__(self, simulation_model, agent_id, origin, dwell_time, seats, **kwargs):
        super().__init__(simulation_model, agent_id, origin, seats, dwell_time=dwell_time, **kwargs)

    def loop_(self):
        while True:
            # be idle if planning is empty
            yield self.execute_process(self.test_idle_())

            next_stop = self.planning[0]

            # move to next stop
            self.tempDestination = next_stop.position
            yield self.execute_process(self.move_())

            # wait for the fixed processing time
            if (
                next_stop.arrivalTime is not None
                and next_stop.arrivalTime > self.sim.scheduler.now()
            ):
                yield self.execute_process(
                    self.wait_(next_stop.arrivalTime - self.sim.scheduler.now())
                )

            # process the stop
            yield self.execute_process(self.process_stop_(next_stop))



from starling_sim.model_simulator import launch_simulation
from starling_sim.utils.utils import gz_decompression
from starling_sim.simulation_scenario import SimulationScenario
from starling_sim.utils import paths

import os
import shutil
import subprocess
import filecmp

#: folder containing test scenarios
SIMULATION_TEST_DATA_FOLDER = "tests/simulation_test_data/"

#: name of the reference folder (next to inputs and outputs)
REFERENCE_OUTPUTS_FOLDER_NAME = "reference"


def get_test_scenarios(models=None):
    """
    Enumerate test scenarios for the given models.

    If models is not provided, enumerate scenarios
    for all models available.

    /!\ It is expected that the data folder variable
    paths._DATA_FOLDER has already been modified to
    designate the test data folder.

    :param models: list of model codes

    :return: list of (model_code, scenario) models
    """

    # get list of tested models
    if models is not None and len(models) > 0:
        tested_models = models
    else:
        # get the list of models with a test folder
        tested_models = os.listdir(paths.models_folder())

    # get list of test scenarios
    # store a list of (model, scenario) tuples
    test_scenarios = []
    for model_code in tested_models:
        scenarios = os.listdir(paths.model_folder(model_code))
        for scenario in scenarios:
            test_scenarios.append((model_code, scenario))

    return test_scenarios


def run_model_test(model, scenario, pkg="starling_sim"):
    """
    Run the given test scenario and compare outputs.

    Outputs are compared to reference files stored in the scenario folder.

    :param model: model code
    :param scenario: scenario folder name
    :param pkg: starling package (see --help or run.py)

    :return: boolean indicating if test was successful
    :raises: ValueError if a problem occurs
    """

    # get scenario paths
    model_folder_path = paths.model_folder(model)
    scenario_path = os.path.join(model_folder_path, scenario, "")
    simulation_scenario = SimulationScenario(scenario_path)

    # remove existing outputs
    output_folder = simulation_scenario.outputs_folder
    if os.path.exists(output_folder):
        shutil.rmtree(output_folder)

    try:
        # run the scenario
        launch_simulation(scenario_path, pkg)
    except Exception as e:
        message = "Simulation crash ({})".format(str(e))
        raise ValueError(message)

    # compare the scenario outputs with reference files
    compare_scenario_outputs(simulation_scenario)

    return True


def compare_scenario_outputs(simulation_scenario):
    """
    Compare test scenario outputs with files of the reference folder.

    :param simulation_scenario: SimulationScenario object
    :raises: ValueError if files don't match
    """

    # get the test files
    test_scenario_output_folder = simulation_scenario.outputs_folder
    test_scenario_output_files = os.listdir(test_scenario_output_folder)

    # extract bz2 and gz archives
    for output_file in test_scenario_output_files:
        if output_file.endswith(".bz2"):
            subprocess.run(["bzip2", "-d", "-f", test_scenario_output_folder + output_file])
        if output_file.endswith(".gz"):
            gz_decompression(test_scenario_output_folder + output_file)

    # get the reference files
    scenario_expected_outputs_folder = os.path.join(
        simulation_scenario.scenario_folder, REFERENCE_OUTPUTS_FOLDER_NAME, ""
    )
    expected_output_files_list = os.listdir(scenario_expected_outputs_folder)

    # compare the files
    match, mismatch, errors = filecmp.cmpfiles(
        test_scenario_output_folder, scenario_expected_outputs_folder, expected_output_files_list
    )

    # raise errors if mismatch or error

    if len(mismatch) != 0:
        raise ValueError("Mismatches in the output files {}".format(mismatch))

    if len(errors) != 0:
        raise ValueError("Errors in the comparison of files {}".format(errors))



from starling_sim.basemodel.agent.persons.person import Person


class User(Person):
    """
    This class describes a simple station-based vehicle-sharing user
    """

    SCHEMA = {
        "properties": {
            "has_station_info": {
                "title": "Station stock information",
                "description": "Indicate if user has access to vehicle availability of the stations",
                "type": "boolean",
                "default": True,
            },
            "patience": {
                "advanced": True,
                "title": "Patience when requesting a station",
                "description": "user patience while waiting for a vehicle. "
                "Caution, None means infinite patience",
                "type": ["integer", "null"],
                "minimum": 0,
                "default": None,
            },
            "closest_station_evaluation": {
                "advanced": True,
                "title": "Station distance evaluation",
                "description": "Determine how the distance to the stations is evaluated",
                "type": "string",
                "oneOf": [{"const": "euclidean"}, {"const": "shortest_path"}],
                "default": "euclidean",
            },
        },
        "required": ["has_station_info", "closest_station_evaluation"],
    }

    def __init__(self, simulation_model, agent_id, origin, destination, **kwargs):
        super().__init__(simulation_model, agent_id, origin, destination, **kwargs)

        # remember the failed stations
        self.failed_stations_ids = []

    def loop_(self):
        """
        Main loop of a station-based vehicle-sharing user
        """

        # loop on trying to get vehicle at closest station
        yield self.execute_process(self.request_loop_(self.maxTries))

        if self.vehicle is None:
            # if failed to get a vehicle, leave the system
            self.leave_simulation("FAIL_GET")
        else:
            # loop on trying to leave vehicle at station closest to dest
            yield self.execute_process(self.request_loop_())

            # should not have a vehicle
            if self.vehicle is not None:
                self.simulation_error("Did not return its vehicle")
            else:
                # end trip
                yield self.execute_process(self.walk_to_destination_())

            return

    def try_new_request_(self):
        """
        Try to request the best station according to user's state.

        The best station is the closest to agent. According to their
        has_station_info attribute, users may ignore the stations that can't
        offer them the service they look for. If no station is found, they wait
        30 seconds and try again.

        :return: StationRequest object, completed according to request result
        """

        # find the best station according to current state
        best_station = self.best_station_for()

        while best_station is None:
            yield self.execute_process(self.spend_time_(30))
            best_station = self.best_station_for()

        # go to chosen station and request it
        self.tempDestination = best_station.position

        if self.vehicle is None:
            # GET request
            yield self.execute_process(self.move_())
            request = self.station_get_request(best_station)
            yield self.execute_process(self.request_tries(request))

            if request.success:
                self.get_vehicle(request.result)
                self.failed_stations_ids = []
            else:
                self.failed_stations_ids.append(best_station.id)

        else:
            # PUT request
            yield self.execute_process(self.move_shortest_with_vehicle_())
            request = self.station_put_request(best_station)
            yield self.execute_process(self.request_tries(request))

            if request.success:
                self.leave_vehicle()
            else:
                self.failed_stations_ids.append(best_station.id)

        # return request to loop
        return request

    def request_tries(self, station_request):
        """
        Execute a sequence of attempts for the given request.

        Each time the waiting time exceeds the agent's patience,
        they reevaluate their decision of staying in the queue.
        This evaluation is done by the quit_decision method.
        This results in a Request with a sequence of waiting times

        :param station_request: StationRequest object
        """

        # request event
        wait_sequence = []

        # quit boolean
        quitting = False

        while not quitting:
            start = self.sim.scheduler.now()
            result = yield self.execute_process(
                self.wait_for_request_(station_request, self.profile["patience"])
            )

            # trace wait event
            wait_sequence += [self.sim.scheduler.now() - start]

            # result of the request
            if station_request.event_ in result:
                # complete request
                station_request.waitSequence = wait_sequence
                station_request.success = True
                station_request.result = result[station_request.event_]

                # end request
                station_request.event_.cancel()

                # return result of request
                return
                # return result[station_request.event_]
            else:
                quitting = self.quit_decision(station_request)

        if quitting:
            # complete request
            station_request.waitSequence = wait_sequence
            station_request.success = False

            # end request
            station_request.event_.cancel()

            # return 0 in case of failed request
            # return 0

    def best_station_for(self):
        """
        Return the best station according to the agent's state.

        It corresponds to the closest station to either current position
        or destination, depending on the agent having a vehicle or not.
        Some stations may be ignored (e.g. empty stations), depending on the information
        detained by the agent

        :return: Station object, or None if no station is found
        """

        # compute stations to consider for closest search
        considered_stations = []

        for station in self.sim.agentPopulation["station"].values():
            # don't consider the stations where the request already failed
            if station.id in self.failed_stations_ids:
                continue

            if self.profile["has_station_info"]:
                # don't consider empty stations if looking for a vehicle, and vice-versa
                if (self.vehicle is None and station.nb_products() == 0) or (
                    self.vehicle is not None and station.nb_products() == station.capacity
                ):
                    continue

            considered_stations.append(station)

        # compute closest station to either current position or destination
        if self.profile["closest_station_evaluation"] == "euclidean":
            if self.vehicle is None:
                best_station = self.sim.environment.euclidean_n_closest(
                    self.position, considered_stations, 1
                )[0]
            else:
                best_station = self.sim.environment.euclidean_n_closest(
                    self.destination, considered_stations, 1
                )[0]

        # TODO : return path and use it
        elif self.profile["closest_station_evaluation"] == "shortest_path":
            if self.vehicle is None:
                best_station = self.sim.environment.closest_object(
                    self.position, considered_stations, True, "walk", n=3
                )
            else:
                best_station = self.sim.environment.closest_object(
                    self.destination, considered_stations, False, "walk", n=3
                )
        else:
            raise ValueError(
                "Unsupported value {} for 'closest_station_evaluation' option.".format(
                    self.profile["closest_station_evaluation"]
                )
            )

        return best_station

    def quit_decision(self, request):
        """
        Decide if user should quit after exceeding its request patience

        :param request: StationRequest object
        :return: boolean, True for quitting the request queue
        """

        return True



"""
Test Starling models
"""

from starling_sim.utils.testing import run_model_test


def test_model_scenario(model, scenario):
    """
    Run a test scenario. See utils.testing.py for more information.
    """
    assert run_model_test(model, scenario)
